import logging
from enum import Enum


class LEVELS(int, Enum):
    """
    Proxy the Logging levels so we can just reference these without
    having to import Logging everywhere.

    LEVEL       | PURPOSE                                                        | Format
    ----------- | -------------------------------------------------------------- | ----------------------------------
    DEBUG       | Information for engineers to observe inner state and flow      | Format as desired
    INFO        | Information recording user and system events                   | Messages should be JSON formatted
    WARNING     | Undesirable but workable event has happened                    | Messages should be informative
    ERROR       | A problem has happened that stopped a minor part of the system | Messages should be instructive
    AUDIT       | Information relating to proving activities happened            | Messages should be JSON formatted
    ALERT       | Intervention is required                                       | Messages should be instructive
    """

    DEBUG = int(logging.DEBUG)  # 10
    INFO = int(logging.INFO)  # 20
    WARNING = int(logging.WARNING)  # 30
    ERROR = int(logging.ERROR)  # 40
    AUDIT = 80
    ALERT = 90


LEVELS_TO_STRING = {
    LEVELS.DEBUG: "DEBUG",
    LEVELS.INFO: "INFO",
    LEVELS.WARNING: "WARNING",
    LEVELS.ERROR: "ERROR",
    LEVELS.AUDIT: "AUDIT",
    LEVELS.ALERT: "ALERT",
}



import os
import sys
import numpy

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.compute.compiled import collect_cython

def test_collector():
    columns = collect_cython([(1, 2), (2, 1), (7, 8)], numpy.array([1, 0], dtype=numpy.int32))
    assert len(columns) == 2
    assert len(columns[0]) == 3
    assert sum(columns[0]) == 11
    assert sum(columns[1]) == 10

def test_collector_empty_input():
    columns = collect_cython([], numpy.array([], dtype=numpy.int32))
    assert len(columns) == 0, len(columns)

def test_collector_single_tuple():
    columns = collect_cython([(5, 10)], numpy.array([1], dtype=numpy.int32))
    assert len(columns) == 1, len(columns)
    assert columns[0] == [10]

def test_collector_large_data():
    data = [(i, i * 2) for i in range(10000)]
    index = numpy.array([1, 0], dtype=numpy.int32)
    columns = collect_cython(data, index)
    assert len(columns) == 2
    assert len(columns[0]) == 10000
    assert sum(columns[0]) == sum(i * 2 for i in range(10000))
    assert sum(columns[1]) == sum(range(10000))


def test_collector_non_integer_index():
    data = [(1, 2), (3, 4)]
    index = numpy.array([0.5, 1.5], dtype=numpy.float64)
    try:
        collect_cython(data, index)
        assert False, "Expected a ValueError"
    except ValueError:
        pass

def test_collector_negative_index():
    data = [(1, 2), (3, 4)]
    index = numpy.array([-1, 0], dtype=numpy.int32)
    try:
        collect_cython(data, index)
        assert False, "Expected an IndexError"
    except IndexError:
        pass

def test_collector_large_index_values():
    data = [(1, 2), (3, 4)]
    index = numpy.array([100, 200], dtype=numpy.int32)
    try:
        collect_cython(data, index)
        assert False, "Expected an IndexError"
    except IndexError:
        pass

def test_collector_duplicate_indices():
    data = [(1, 2), (3, 4), (5, 6)]
    index = numpy.array([1, 1, 0], dtype=numpy.int32)
    columns = collect_cython(data, index)
    assert len(columns) == 3
    assert sum(columns[0]) == 12, sum(columns[0])
    assert sum(columns[1]) == 12, sum(columns[1])
    assert sum(columns[2]) == 9, sum(columns[2])

if __name__ == "__main__":  # pragma: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import typing

from orso.compute.compiled import process_table
from orso.exceptions import MissingDependencyError
from orso.row import Row
from orso.schema import FlatColumn
from orso.schema import RelationSchema


class _RowsIterator:
    """
    Iterator class for processing PyArrow tables lazily.

    Parameters:
        tables: Iterable of PyArrow tables to process.
        row_factory: Factory method to create Row instances.
        batch_size: Number of rows to process at a time.
        max_size: Maximum number of rows to return.
    """

    def __init__(
        self, tables: typing.Iterable, row_factory: typing.Callable, batch_size: int, max_size: int
    ):
        self.tables = itertools.chain(tables)
        self.row_factory = row_factory
        self.batch_size = batch_size
        self.max_size = max_size
        self.rows_processed = 0
        self.current_table = None
        self.current_rows = iter([])

    def __iter__(self):
        return self

    def __next__(self):
        if self.rows_processed >= self.max_size:
            raise StopIteration()

        row = next(self.current_rows, None)
        if row is not None:
            self.rows_processed += 1
            return row
        else:
            # Fetch the next table and process it
            self.current_table = next(self.tables, None)
            if self.current_table is None:
                raise StopIteration()

            self.current_rows = iter(
                process_table(self.current_table, self.row_factory, self.batch_size)
            )

            # Check if the new table has rows to process
            row = next(self.current_rows, None)
            if row is not None:
                self.rows_processed += 1
                return row
            else:
                raise StopIteration()


def to_arrow(dataset, size=None):
    try:
        import pyarrow
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error

    if size is not None and size >= 0:
        dataset = dataset.head(size)

    if dataset.rowcount == 0:
        arrays = [list() for _ in range(dataset.columncount)]
    else:
        arrays = list(zip(*dataset._rows))

    return pyarrow.Table.from_arrays(arrays, dataset.column_names)


def from_arrow(tables, size=None):
    """
    Convert an Arrow table or an iterable of Arrow tables to a generator of
    Python objects, handling each block one at a time.
    """
    if not isinstance(tables, (typing.Generator, list, tuple)):
        tables = [tables]

    if isinstance(tables, (list, tuple)):
        tables = iter(tables)

    BATCH_SIZE: int = 10_000
    if size:
        BATCH_SIZE = min(size, BATCH_SIZE)
    else:
        size = float("inf")

    # Extract schema information from the first table
    first_table = next(tables, None)
    if first_table is None:
        return iter([]), {}

    arrow_schema = first_table.schema
    orso_schema = RelationSchema(
        name="arrow",
        columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
    )
    row_factory = Row.create_class(orso_schema, tuples_only=True)

    # Create an bespoke lazy iterator instance
    rows_iterator = _RowsIterator(
        tables=itertools.chain([first_table], tables),
        row_factory=row_factory,
        batch_size=BATCH_SIZE,
        max_size=size,
    )

    return rows_iterator, orso_schema


def to_pandas(dataset, size=None):
    """wrap the arrow function to convert to pandas"""
    return dataset.arrow(size).to_pandas()


def from_pandas(pandas):
    raise NotImplementedError()


def to_polars(dataset, size=None):
    try:
        import polars
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error
    return polars.DataFrame(r.as_dict for r in dataset.slice(0, size))


def from_polars(polars):
    raise NotImplementedError()


def to_csv(dataset):
    raise NotImplementedError()


def from_csv(values):
    raise NotImplementedError()


def to_parquet(dataset):
    raise NotImplementedError()


def from_parquet(values):
    raise NotImplementedError()


def to_jsonl(dataset):
    raise NotImplementedError()


def from_jsonl(values):
    raise NotImplementedError()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy

from orso import Row
from orso.group_by import GroupBy
from orso.schema import RelationSchema
from orso.tools import single_item_cache


class DataFrame:
    """
    Orso DataFrames are a lightweight container for tabular data.
    """

    __slots__ = ("_schema", "_rows", "_cursor", "_row_factory", "arraysize", "_nbytes")

    def __init__(
        self,
        dictionaries: Optional[Iterable[dict]] = None,
        *,
        rows: Union[List[tuple], Generator[tuple, None, None], None] = None,
        schema: Union[RelationSchema, List[str], Tuple[str], None] = None,
    ):
        """
        Create an orso DataFrame. DataFrames are a representation of a table with
        rows of records which have the same fields of the same data types.

        Parameters:
            dictionaries: iterable of dicts (optional)
                An iterable of dictionaries. The schema for the frame is determined
                from the first dictionary in the iterable. Schemas for dictionary
                defined DataFrames are forgiving, missing fields are substituted with
                None and common types are not enforced.
            rows: iterable of tuples (optional, keyworded)
                An iterable of tuples representing a row in the DataFrame. Rows should
                conform to the schema, this is not enforced at creation. Must be used
                with 'schema' and cannot be used with 'dictionaries'.
            schema: RelationSchema or list of column names (optional, keyworded)
                A RelationSchema describing the schema of the DataFrame or a list of
                the column names. Must be used with 'rows' and cannot be used with
                'dictionaries'

        """
        self._nbytes = None
        if dictionaries is not None:
            if schema is not None or rows is not None:
                raise ValueError(
                    "Can't implicitly and explicitly define a DataFrame at the same time"
                )

            from itertools import chain

            # make the list of dicts iterable
            dicts = iter(dictionaries)
            # extract the first of the list, and get the types from it
            first_dict = next(dicts)

            # if we have an explicit schema, use that, otherwise guess from the first entry
            self._schema = [str(k) for k in first_dict]
            self._row_factory = Row.create_class(self._schema)
            keys = list(first_dict.keys())

            # create a list of tuples
            self._rows = [
                self._row_factory([row.get(k, None) for k in keys])
                for row in chain([first_dict], dicts)
            ]
        else:
            self._schema = schema  # type:ignore
            self._rows = rows or []  # type:ignore
            self._row_factory = Row.create_class(self._schema)
            self._nbytes = 0
        self.arraysize = 100
        self._cursor = iter(self._rows or [])

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def group_by(self, columns: List[str]) -> GroupBy:
        return GroupBy(self, columns)

    @classmethod
    def from_arrow(cls, tables):
        """Convert a PyArrow Table, or list of Tables, to an Orso DataFrame."""
        from orso.converters import from_arrow

        rows, schema = from_arrow(tables)
        return cls(rows=rows, schema=schema)

    def arrow(self, size=None):
        """
        Convert an Orso DataFrame to a PyArrow Table, optionally limit the number
        of records.
        """
        from orso.converters import to_arrow

        return to_arrow(self, size=size)

    def pandas(self, size=None):
        from orso.converters import to_pandas

        return to_pandas(self, size)

    def polars(self, size=None):
        from orso.converters import to_polars

        return to_polars(self, size)

    def nbytes(self) -> int:
        """Approximate the number of bytes used by the DataFrame"""
        self.materialize()
        if self._nbytes is None:
            self._nbytes = sum(row.nbytes() for row in self._rows)
        return self._nbytes

    def append(self, entry):
        if isinstance(self._schema, RelationSchema):
            self._schema.validate(entry)
        new_row = self._row_factory(entry)
        self._rows.append(new_row)
        self._nbytes += new_row.nbytes()
        self._cursor = None

    def head(self, size: int = 5) -> "DataFrame":
        return self.slice(0, size)

    def tail(self, size: int = 5) -> "DataFrame":
        return self.slice(offset=0 - size, length=size)

    def query(self, predicate) -> "DataFrame":
        """
        Apply a Selection operation to a Relation, this filters the data in the
        Relation to just the entries which match the predicate.

        Parameters:
            predicate (callable):
                A function which can be applied to a tuple to determine if it
                should be returned in the target Relation.

        Returns:
            DataFrame
        """
        return DataFrame(rows=[r for r in self._rows if predicate(r)], schema=self._schema)

    def select(self, attributes) -> "DataFrame":
        """
        Create a new Orso DataFrame with a subset of columns of an existing DataFrame
        """
        if not isinstance(attributes, (list, tuple)):
            attributes = [attributes]
        attribute_indices = []
        new_header = attributes
        for index, attribute in enumerate(self._schema):
            if attribute in attributes:
                attribute_indices.append(index)

        def _inner_projection():
            for tup in self._rows:
                yield tuple([tup[indice] for indice in attribute_indices])

        return DataFrame(rows=_inner_projection(), schema=new_header)

    def materialize(self):
        """
        Convert a Lazy DataFrame to an Eager DataFrame
        """
        if not isinstance(self._rows, list):
            self._rows = list(self._rows or [])

    def distinct(self) -> "DataFrame":
        seen = set()
        unique_rows = [
            x
            for x in self._rows
            if hash(x) not in seen and not seen.add(hash(x))  # type:ignore
        ]
        return DataFrame(rows=unique_rows, schema=self._schema)

    def collect(
        self, columns: Union[int, str, List[Union[int, str]]], limit: int = None
    ) -> Union[List, Tuple]:
        """
        Collects specified columns from the internal row storage into a tuple of lists.

        Parameters:
            columns: Union[int, str, List[Union[int, str]]]
                The column(s) to collect. Could be an integer index, string name, or list thereof.
            limit: int (optional)
                The number of rows to return, defaults to all rows

        Returns:
            Union[list, tuple]:
                A tuple containing lists of the column data, or a single list if only one column is specified.
        """
        from orso.compute.compiled import collect_cython

        self.materialize()

        if limit is None or limit < 0:
            limit = -1

        single = False
        if not isinstance(columns, (list, set, tuple)):
            single = True
            columns = [columns]
        else:
            columns = list(columns)

        column_indicies = columns
        for i, c in enumerate(columns):
            if not isinstance(c, int):
                column_indicies[i] = self.column_names.index(c)

        collected = collect_cython(
            self._rows, numpy.array(column_indicies, dtype=numpy.int32), limit
        )
        if single:
            return collected[0]
        return collected

    def __getitem__(self, items):
        return self.collect(columns=items, limit=None)

    def slice(self, offset: int = 0, length: int = None) -> "DataFrame":
        self.materialize()
        if offset < 0:
            offset = len(self._rows) + offset
        if length is None:
            return DataFrame(schema=self._schema, rows=self._rows[offset:])
        if length == 0:
            return DataFrame(schema=self._schema, rows=[])
        return DataFrame(schema=self._schema, rows=self._rows[offset : offset + length])

    def filter(self, mask) -> "DataFrame":
        """
        Select rows from the DataFRame. The DataFrame is filtered based on boolean array.
        """
        return DataFrame(schema=self._schema, rows=(t for t, m in zip(self._rows, mask) if m))

    def take(self, indexes) -> "DataFrame":
        """
        Select rows from the DataFrame. Rows are selected based on their appearance in the indexes
        list
        """
        return DataFrame(
            schema=self._schema, rows=(m for i, m in enumerate(self._rows) if i in indexes)
        )

    def row(self, i) -> Row:
        self.materialize()
        return self._rows[i]

    def fetchone(self) -> Row:
        if self._cursor is None:
            raise Exception("Cannot use fetchone and append on the same DataFrame")
        try:
            return next(self._cursor)
        except StopIteration:
            return None

    def fetchmany(self, size=None) -> List[Row]:
        if self._cursor is None:
            raise Exception("Cannot use fetchmany and append on the same DataFrame")
        fetch_size = self.arraysize if size is None else size
        entries = []
        for i in range(fetch_size):
            try:
                entry = next(self._cursor)
                entries.append(entry)
            except StopIteration:
                break
        return entries

    def fetchall(self) -> List[Row]:
        if self._cursor is None:
            raise Exception("Cannot use fetchall and append on the same DataFrame")
        return list(self._cursor)

    def display(
        self,
        limit: int = 5,
        display_width: Union[bool, int] = True,
        max_column_width: int = 32,
        colorize: bool = True,
        show_types: bool = True,
    ) -> str:
        from .display import ascii_table

        return ascii_table(
            self,
            limit=limit,
            display_width=display_width,
            max_column_width=max_column_width,
            colorize=colorize,
            show_types=show_types,
        )

    def markdown(self, limit: int = 5, max_column_width: int = 30) -> str:
        from .display import markdown

        return "\n".join(markdown(self, limit=limit, max_column_width=max_column_width))

    def to_batches(self, batch_size: int = 1000) -> typing.Generator["DataFrame", None, None]:
        """
        Batch a DataFrame into batches of `size` records.

        Args:
            size (int): The size of each batch.

        Yields:
            DataFrames
        """
        self.materialize()
        for i in range(0, self.rowcount, batch_size):
            yield DataFrame(rows=self._rows[i : i + batch_size], schema=self._schema)

    @property
    def profile(self) -> "TableProfile":
        from orso.profiler import TableProfile

        return TableProfile.from_dataframe(self)

    @property
    def description(
        self,
    ) -> List[
        Tuple[
            str,
            Optional[str],
            None,
            None,
            Optional[int],
            Optional[int],
            Optional[bool],
        ]
    ]:
        """
        name
        type_code
        display_size
        internal_size
        precision
        scale
        null_ok
        """
        result = []
        for column in self.column_names:
            data_type = None
            data_precision = None
            data_scale = None
            nullable = None
            if isinstance(self._schema, RelationSchema):
                column_data = self._schema.find_column(column)
                column_type = column_data.type
                if column_type is not None:
                    data_type = str(column_data.type.value)
                if column_type.value == "DECIMAL":
                    data_precision = column_data.precision
                    data_scale = column_data.scale
                    data_type = f"DECIMAL({data_precision},{data_scale})"
                if column_type.value == "ARRAY" and column_data.element_type is not None:
                    data_type = f"ARRAY<{column_data.element_type.value}>"
                nullable = column_data.nullable
            result.append(
                (
                    column,
                    data_type,
                    None,
                    None,
                    data_precision,
                    data_scale,
                    nullable,
                )
            )
        return result  # type:ignore

    @property
    @single_item_cache
    def column_names(self) -> Tuple[str]:
        if isinstance(self._schema, (tuple, list)):
            return tuple(str(c) for c in self._schema)  # type:ignore
        return tuple(str(col.name) for col in self._schema.columns)

    @property
    @single_item_cache
    def columncount(self) -> int:
        if isinstance(self._schema, (tuple, list)):
            return len(self._schema)
        return len(self._schema.columns)

    @property
    def shape(self) -> typing.Tuple[int, int]:
        return (self.rowcount, self.columncount)  # type: ignore

    @property
    def rowcount(self) -> int:
        self.materialize()
        return len(self._rows)

    @property
    def schema(self) -> RelationSchema:
        return self._schema

    def __hash__(self):
        from xxhash import xxh64

        _hash = 0
        for i, row in enumerate(self._rows):
            row_hash = xxh64(str(row).encode()).intdigest()
            _hash = i ^ _hash ^ row_hash
        return _hash

    def __iter__(self):
        return iter(self._rows)

    def __len__(self) -> int:
        self.materialize()
        return len(self._rows)

    def __repr__(self) -> str:
        """
        If we're in a notebook, return a table, otherwise return a tag
        """
        try:
            from IPython import get_ipython

            i_am_in_a_notebook = get_ipython() is not None
        except ImportError:
            return f"<orso.dataframe>"
        return str(self)

    def __str__(self) -> str:
        size: int = 10
        try:
            from IPython import get_ipython

            i_am_in_a_notebook = get_ipython() is not None
        except ImportError:
            i_am_in_a_notebook = False

        if i_am_in_a_notebook:
            from IPython.display import HTML
            from IPython.display import display

            from .display import html_table

            self.materialize()
            html = html_table(self, size)
            display(HTML(html))
            return ""
        else:
            from .display import ascii_table

            return (
                ascii_table(self, limit=size, top_and_tail=True)
                + f"\n[ {self.rowcount} rows x {self.columncount} columns ]"
            )

    def __add__(self, the_other):
        if self._schema != the_other._schema:
            raise ValueError("Schemas must be identical to add DataFrames")
        return DataFrame(rows=self._rows + the_other._rows, schema=self._schema)



import datetime
import decimal
import pytest
import sys


sys.path.insert(1, "/".join([str(p) for p in __file__.split("/")[:-2]])) 

print(sys.path)

from orso.types import OrsoTypes
import orso

print(orso.__version__)

CAST_TESTS = [
    ("BOOLEAN", "true", True),
    ("BOOLEAN", "FALSE", False),
    ("BOOLEAN", None, None),
    ("BOOLEAN", "yes", True),
    ("BOOLEAN", "no", False),
    ("BOOLEAN", "on", True),
    ("BOOLEAN", "off", False),
    ("BOOLEAN", "1", True),
    ("BOOLEAN", "0", False),
    ("BOOLEAN", 1, True),
    ("BOOLEAN", 0, False),
    ("BOOLEAN", 1.000, True),
    ("BOOLEAN", 0.000, False),
    ("BOOLEAN", True, True),
    ("BOOLEAN", False, False),
    ("BOOLEAN", "y", True),
    ("BOOLEAN", "n", False),
    ("BOOLEAN", "t", True),
    ("BOOLEAN", "f", False),
    ("BOOLEAN", b"true", True),
    ("BOOLEAN", b"FALSE", False),
    ("BOOLEAN", b"yes", True),
    ("BOOLEAN", b"no", False),
    ("BOOLEAN", b"on", True),
    ("BOOLEAN", b"off", False),
    ("BOOLEAN", b"1", True),
    ("BOOLEAN", b"0", False),
    ("BOOLEAN", 1, True),
    ("BOOLEAN", 0, False),
    ("BOOLEAN", 1.000, True),
    ("BOOLEAN", 0.000, False),
    ("BOOLEAN", True, True),
    ("BOOLEAN", False, False),
    ("BOOLEAN", b"y", True),
    ("BOOLEAN", b"n", False),
    ("BOOLEAN", b"t", True),
    ("BOOLEAN", b"f", False),

    ("BLOB", "hello", b"hello"),
    ("BLOB", b"bytes", b"bytes"),
    ("BLOB", None, None),
    ("BLOB[5]", "hello world", b"hello"),  # Test length limit
    ("BLOB", 123, b"123"),
    ("BLOB", 123.45, b"123.45"),
    ("BLOB", True, b"True"),
    ("BLOB", False, b"False"),
    ("BLOB", "", b""),
    ("BLOB", b"", b""),
    ("BLOB", "特殊字符", b"\xe7\x89\xb9\xe6\xae\x8a\xe5\xad\x97\xe7\xac\xa6"),
    ("BLOB", ["list", "item"], b'["list","item"]'),
    ("BLOB", {"key": "value"}, b'{"key":"value"}'),
   
    ("DATE", "2023-04-18", datetime.date(2023, 4, 18)),
    ("DATE", b"2023-04-18", datetime.date(2023, 4, 18)),
    ("DATE", "2023-04-18 12:12:12", datetime.date(2023, 4, 18)),
    ("DATE", datetime.date(2023, 4, 18), datetime.date(2023, 4, 18)),
    ("DATE", datetime.datetime(2023, 4, 18, 12, 34, 56), datetime.date(2023, 4, 18)),
    ("DATE", 1681776000, datetime.date(2023, 4, 18)),  # Unix timestamp for 2023-04-18
    ("DATE", None, None),
    ("DATE", "2024-02-29", datetime.date(2024, 2, 29)),  # Valid leap year date

    ("TIMESTAMP", "2023-04-18T12:34:56", datetime.datetime(2023, 4, 18, 12, 34, 56)),
    ("TIMESTAMP", b"2023-04-18T12:34:56", datetime.datetime(2023, 4, 18, 12, 34, 56)),
    ("TIMESTAMP", None, None),
    ("TIMESTAMP", "2023-04-18", datetime.datetime(2023, 4, 18, 0, 0, 0)),
    ("TIMESTAMP", 1681776600, datetime.datetime(2023, 4, 18, 0, 10, 0)),

    ("DECIMAL", "123.45", decimal.Decimal("123.45")),
    ("DECIMAL", 123, decimal.Decimal(123)),
    ("DECIMAL(5,3)", "123.45", decimal.Decimal("123.45")),
    ("DECIMAL(4,4)", "123.45", decimal.Decimal("123.4")),
    ("DECIMAL(5,3)", b"123.45", decimal.Decimal("123.45")),

    ("DOUBLE", "123.45", 123.45),
    ("DOUBLE", 123, 123.0),
    ("DOUBLE", "-123.45", -123.45),
    ("DOUBLE", b"123.45", 123.45),
    ("DOUBLE", "1.23e5", 123000.0),
    ("DOUBLE", "1.23E-5", 0.0000123),
    ("DOUBLE", float('inf'), float('inf')),
    ("DOUBLE", "-inf", float('-inf')),
    ("DOUBLE", None, None),
    ("DOUBLE", True, 1.0),
    ("DOUBLE", False, 0.0),
    ("DOUBLE", " 123.45 ", 123.45),  # Test trimming
    ("DOUBLE", "000123.45000", 123.45),  # Test trimming

    ("INTEGER", "42", 42),
    ("INTEGER", 42.9, 42),
    ("INTEGER", "-42", -42),
    ("INTEGER", b"42", 42),
    ("INTEGER", True, 1),
    ("INTEGER", False, 0),
    ("INTEGER", None, None),
    ("INTEGER", "0042", 42),  # Test leading zeros
    ("INTEGER", " 42 ", 42),  # Test trimming
    ("INTEGER", 9223372036854775807, 9223372036854775807),  # Max int64
    ("INTEGER", -9223372036854775808, -9223372036854775808),  # Min int64

    ("VARCHAR", 123, "123"),
    ("VARCHAR", "hello", "hello"),
    ("VARCHAR", None, None),
    ("VARCHAR", 123.45, "123.45"),
    ("VARCHAR", True, "True"),
    ("VARCHAR", False, "False"),
    ("VARCHAR", b"binary string", "binary string"),
    ("VARCHAR", ["list", "items"], "['list', 'items']"),
    ("VARCHAR", {"key": "value"}, "{'key': 'value'}"),
    ("VARCHAR", datetime.date(2023, 4, 18), "2023-04-18"),
    ("VARCHAR", datetime.datetime(2023, 4, 18, 12, 34, 56), "2023-04-18 12:34:56"),
    ("VARCHAR", "", ""),
    ("VARCHAR[5]", "hello world", "hello"),  # Test length limit
    ("VARCHAR[10]", "hello", "hello"),  # Test length under limit
    ("VARCHAR", " padded ", " padded "),  # Test that spaces are preserved

    ("NULL", "anything", None),
    ("NULL", None, None),

    ("ARRAY<INTEGER>", [1, 2, 3], [1, 2, 3]),
    ("ARRAY<INTEGER>", "[1, 2, 3]", [1, 2, 3]),
    ("ARRAY<INTEGER>", b"[1, 2, 3]", [1, 2, 3]),
    ("ARRAY<INTEGER>", "[1, null, 3]", [1, None, 3]),
    ("ARRAY<INTEGER>", None, None),
    ("ARRAY<INTEGER>", [], []),

    ("ARRAY<DOUBLE>", [1.1, 2.2, 3.3], [1.1, 2.2, 3.3]),
    ("ARRAY<DOUBLE>", "[1.1, 2.2, 3.3]", [1.1, 2.2, 3.3]),

    ("ARRAY<BOOLEAN>", [True, False, True], [True, False, True]),
    ("ARRAY<BOOLEAN>", '["true", "false", "yes"]', [True, False, True]),

    ("ARRAY<VARCHAR>", ["a", "b", "c"], ["a", "b", "c"]),
    ("ARRAY<VARCHAR>", [1, 2, 3], ["1", "2", "3"]),

    ("ARRAY<BLOB>", '["hello", "world"]', [b"hello", b"world"]),
    ("ARRAY<BLOB>", b'["hello", "world"]', [b"hello", b"world"]),

    ("ARRAY<DATE>", ["2023-04-18", "2024-02-29"], [datetime.date(2023, 4, 18), datetime.date(2024, 2, 29)]),

    ("ARRAY<TIMESTAMP>", ["2023-04-18T12:34:56", "2023-04-19T00:00:00"], [datetime.datetime(2023, 4, 18, 12, 34, 56), datetime.datetime(2023, 4, 19, 0, 0, 0)]),
]

@pytest.mark.parametrize("type_name,input_value,expected", CAST_TESTS)
def test_orso_type_parsers(type_name, input_value, expected):
    _type, _length, _precision, _scale, _element_type = OrsoTypes.from_name(type_name)
    value = _type.parse(input_value, length=_length, precision=_precision, scale=_scale, element_type=_element_type)
    assert value == expected, f"{type_name} parsing of {input_value} returned {value}, expected {expected}"



BOOLEAN_STRINGS = ("TRUE", "ON", "YES", "1", "1.0", b"TRUE", b"ON", b"YES", b"1", b"1.0")
bp = lambda x: str(x).upper() in BOOLEAN_STRINGS

print(bp("FALSE"))


if __name__ == "__main__":  # pragma: no cover
    print(f"RUNNING BATTERY OF {len(CAST_TESTS)} DATE TESTS")
    for type_name, input_value, expected in CAST_TESTS:
        print(f"{type_name} {input_value} => {expected}" )
        test_orso_type_parsers(type_name, input_value, expected)
    print("okay")





# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import decimal
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union


# Aggregation Functions
def min_agg(values: List[Any]) -> Any:
    return min(values)


def max_agg(values: List[Any]) -> Any:
    return max(values)


def count_agg(values: List[Any]) -> int:
    return len(values)


def avg_agg(values: List[decimal.Decimal]) -> decimal.Decimal:
    return decimal.Decimal(sum(values)) / decimal.Decimal(len(values))


def sum_agg(values: List[decimal.Decimal]) -> decimal.Decimal:
    return sum(values)


AGGREGATORS = {"MIN": min_agg, "MAX": max_agg, "COUNT": count_agg, "AVG": avg_agg, "SUM": sum_agg}


class TooManyGroups(Exception):
    pass


class GroupBy:
    """
    GroupBy does a lazy evaluation of the groups, the groups are calculated as part of
    calculating the aggregations. This was implemented like this so that generators
    can be aggregated - we have one opportunity to cycle of the records, and if the
    data is in a generator, there's a chance the dataset doesn't fit in memory.
    """

    def __init__(self, dictset, columns):
        self._dictset = dictset
        if isinstance(columns, (list, set, tuple)):
            self._columns = tuple(columns)
        else:
            self._columns = [columns]
        self._group_keys = {}

    def _map(self, collect_columns: Union[str, List[str]]) -> Dict[int, Dict[str, List[Any]]]:
        """
        Maps the dataset into groups based on given columns.

        Parameters:
            collect_columns: Union[str, List[str]]
                Columns used for aggregation.

        Returns:
            A dictionary containing groups and aggregated values.
        """
        collect_columns = (
            collect_columns if isinstance(collect_columns, list) else [collect_columns]
        )
        source_columns = self._dictset.column_names
        collect_column_indicies = [
            source_columns.index(target) if target in source_columns else -1
            for target in collect_columns
        ]
        group_column_indicies = array.array(
            "i",
            (source_columns.index(target) for target in self._columns),
        )

        for record in self._dictset:
            # Create a unique hash for each group
            group_key = hash(tuple(record[col] for col in group_column_indicies))

            if group_key not in self._group_keys:
                self._group_keys[group_key] = [
                    (source_columns[column], record[column]) for column in group_column_indicies
                ]

            for i, column in enumerate(collect_column_indicies):
                yield (group_key, collect_columns[i], "*" if column == -1 else record[column])

    def aggregate(self, aggregations: List[Tuple[str, Callable]]) -> "DataFrame":
        """
        Aggregates the grouped data based on the given aggregation functions.

        Parameters:
            aggregations: List[Tuple[str, Callable]]
                A list of tuples containing column name and aggregation function.

        Returns:
            A dictionary containing groups and aggregated values.
        """
        aggregated_data = {}
        column_value_map = defaultdict(lambda: defaultdict(list))

        if not isinstance(aggregations, list):
            aggregations = [aggregations]
        if not all(isinstance(agg, tuple) for agg in aggregations):  # pragma: no cover
            raise ValueError("`aggregate` expects a list of Tuples")

        # Collecting the values for each group and column
        for group_key, column, value in self._map([col for _, col in aggregations]):
            if value is not None:
                column_value_map[group_key][column].append(value)

        # Applying aggregation functions
        for group, column_values in column_value_map.items():
            aggregated_data[group] = {}
            for func, col in aggregations:
                aggregated_data[group][f"{func}({col})"] = AGGREGATORS[func](
                    column_values.get(col, [])
                )

        result_set = []
        for group, values in aggregated_data.items():
            results = {f"{func}({col})": values.get(f"{func}({col})") for func, col in aggregations}
            keys = self._group_keys[group]
            for key in keys:
                results[key[0]] = key[1]
            result_set.append(results)

        from orso.dataframe import DataFrame

        return DataFrame(result_set)

    def max(self, columns) -> "DataFrame":
        """
        Get the maximum value of a column, or set of columns, in each group.

        Parameters:
            columns: string or iterable
                The columns to collect the maximum value of for each group.

        Yields:
            Dictionary
        """
        if not isinstance(columns, (tuple, list, set)):
            columns = [columns]
        return self.aggregate([("MAX", column) for column in columns])

    def min(self, columns) -> "DataFrame":
        """
        Get the minimum value of a column, or set of columns, in each group.

        Parameters:
            columns: string or iterable
                The columns to collect the minimum value of for each group.

        Yields:
            Dictionary
        """
        if not isinstance(columns, (tuple, list, set)):
            columns = [columns]
        return self.aggregate([("MIN", column) for column in columns])

    def sum(self, columns) -> "DataFrame":
        """
        Get the sum of values in a column, or set of columns, in each group.

        Parameters:
            columns: string or iterable
                The columns to calculate the sum of for each group.

        Yields:
            Dictionary
        """
        if not isinstance(columns, (tuple, list, set)):
            columns = [columns]
        return self.aggregate([("SUM", column) for column in columns])

    def count(self) -> "DataFrame":
        """
        Count the number of items in each group.

        Yields:
            Dictionary
        """
        # COUNT is a little different, it doesn't have any fields to perform the
        # aggregation on.
        # This implementation could be improved by taking a copy of the
        # aggregate() function and removing the bits that aren't needed to just
        # count the values.
        return self.aggregate(
            (
                "COUNT",
                "*",
            )
        )

    def avg(self, columns) -> "DataFrame":
        """
        Calculate the average of the items in a group.
        """
        if not isinstance(columns, (tuple, list, set)):
            columns = [columns]
        return self.aggregate([("AVG", column) for column in columns])

    def groups(self) -> "DataFrame":
        """
        Return the set of groups - this is similar to a DISTINCT function
        """
        collector = defaultdict(dict)
        for record in self._map("*"):
            collector[record[0]] = 1

        from orso.dataframe import DataFrame

        return DataFrame(dict(self._group_keys[group]) for group in self._group_keys)



import os
import sys


sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso import logging


def test_logger():
    logger = logging.get_logger()

    logger.error("An Error Occurred")


def test_google_logger():
    logger = logging.google_cloud_logger.GoogleLogger()

    logger.write_event("Google", "Test")


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import hashlib
import json
import logging
import os
import re
from typing import Dict

from orso.display import colorizer

# if we find a key which matches these strings, we hash the contents
KEYS_TO_SANITIZE = [
    r"password$",
    r"pwd$",
    r".*_secret$",
    r".*_key$",
    r"_token$",
    r"credentials",
]
COMPILED_KEYS_TO_SANITIZE = [
    re.compile(expression, re.IGNORECASE) for expression in KEYS_TO_SANITIZE
]

COLOR_EXCHANGES = {
    " ALERT    ": "\001BOLD_REDm ALERT    \001OFFm",
    " ERROR    ": "\001REDm ERROR    \001OFFm",
    " DEBUG    ": "\001GREENm DEBUG    \001OFFm",
    " AUDIT    ": "\001YELLOWm AUDIT    \001OFFm",
    " WARNING  ": "\001BOLD_REDm WARNING  \001OFFm",
    " INFO     ": "\001BOLD_WHITEm INFO     \001OFFm",
}

COLOR_CODES = {
    "KEY": "\001KEYm",
    "OFF": "\001OFFm",
    "PURPLE": "\001PURPLEm",
    "YELLOW": "\001YELLOWm",
    "VALUE": "\001VALUEm",
}

QUOTES_OR_BACKTICKS_RE = re.compile(r"(['`])(.*?)\1")


class LogFormatter(logging.Formatter):
    def __init__(self, orig_formatter, suppress_color: bool = False):
        """
        Remove sensitive data from records before saving to external logs. Note that
        the value is hashed using (SHA256) and only the first 8 characters of the
        hex-encoded hash are presented. This information allows values to be traced
        without disclosing the actual value.

        The Sanitizer can only sanitize dictionaries, it doesn't sanitize strings,
        which could contain sensitive information We use the message id as a salt to
        further protect sensitive information.

        Based On: https://github.com/joocer/cronicl/blob/main/cronicl/utils/sanitizer.py
        """
        self.orig_formatter = orig_formatter
        self.suppress_color = suppress_color

    def format(self, record):
        try:
            msg = self.orig_formatter.format(record)
        except:
            msg = record
        msg = self.sanitize_record(msg)
        if "://" in msg:
            msg = re.sub(r":\/\/(.*?)\@", r"://\001BOLD_PURLEm<redacted>\001OFFm", msg)
        return msg

    def _can_colorize(self) -> bool:
        """
        Determines if the environment supports colorization.

        Returns:
            bool: True if colorization is supported, False otherwise.
        """
        if self.suppress_color:
            return False

        colorterm = os.environ.get("COLORTERM", "").lower()
        term = os.environ.get("TERM", "").lower()
        return "yes" in colorterm or "true" in colorterm or "256" in term

    def color_code(self, record):
        if self._can_colorize():
            for k, v in COLOR_EXCHANGES.items():
                if k in record:
                    return record.replace(k, v)
        return record

    def __getattr__(self, attr):
        return getattr(self.orig_formatter, attr)

    def hash_it(self, value_to_hash: str) -> str:
        """
        Hash a value using SHA256.

        Parameters:
            value_to_hash: str
                The value to hash.

        Returns:
            str: The hashed value.
        """
        return hashlib.sha256(value_to_hash.encode()).hexdigest()[:8]

    def clean_record(self, dirty_record: Dict, colorize: bool = True) -> Dict:
        """
        Cleans a log record dictionary.

        Parameters:
            dirty_record: Dict
                The original log record.
            colorize: bool
                Whether to add color codes to the log.

        Returns:
            Dict: The cleaned log record.
        """
        colors = COLOR_CODES if colorize else {key: "" for key in COLOR_CODES}

        def color_value(match):
            return f"{match.group(1)}{colors['YELLOW']}{match.group(2)}{colors['VALUE']}{match.group(1)}"

        clean_record = {}
        for key, value in dirty_record.items():
            if isinstance(value, dict):
                value = self.clean_record(value, colorize)
            elif any(regex.match(key) for regex in COMPILED_KEYS_TO_SANITIZE):
                value = f"{colors['PURPLE']}<redacted:{self.hash_it(str(value))}>{colors['OFF']}"
            else:
                value = QUOTES_OR_BACKTICKS_RE.sub(color_value, str(value))

            clean_key = f"{colors['KEY']}{key}{colors['OFF']}"
            clean_record[clean_key] = f"{colors['VALUE']}{value}{colors['OFF']}"

        return clean_record

    def sanitize_record(self, record: str) -> str:
        """
        Sanitizes and optionally colorizes a log record.

        Parameters:
            record: str
                The original log record.

        Returns:
            str: The sanitized and optionally colorized log record.
        """
        record = self.color_code(record)
        parts = record.split("|")
        json_part = parts.pop()

        try:
            dirty_record = json.loads(json_part.encode("UTF8"))
            clean_record = self.clean_record(dirty_record)
            parts.append(" " + json.dumps(clean_record))

        except ValueError:
            json_part = re.sub(r"`([^`]*)`", r"`\001YELLOWm\1\001OFFm`", f"{json_part}")
            json_part = re.sub(r"'([^']*)'", r"'\001YELLOWm\1\001OFFm'", f"{json_part}")
            json_part = re.sub(r'"([^"]*)"', r"'\001YELLOWm\1\001OFFm'", f"{json_part}")
            parts.append(" " + json_part.strip() + " *")

        record = "|".join(parts)
        return colorizer(record, self._can_colorize())



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.compute.compiled import extract_dict_columns
import random

def test_extract_dict_columns_basic():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ('a', 'b', 'c')
    result = extract_dict_columns(data, fields)
    assert result == (1, 2, 3), result

def test_extract_dict_columns_missing_fields():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ('a', 'd', 'c')
    result = extract_dict_columns(data, fields)
    assert result == (1, None, 3)

def test_extract_dict_columns_empty_data():
    data = {}
    fields = ('a', 'b', 'c')
    result = extract_dict_columns(data, fields)
    assert result == (None, None, None)

def test_extract_dict_columns_empty_fields():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ()
    result = extract_dict_columns(data, fields)
    assert result == ()

def test_extract_dict_columns_all_fields_missing():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ('x', 'y', 'z')
    result = extract_dict_columns(data, fields)
    assert result == (None, None, None)

def test_extract_dict_columns_large_dataset():
    data = {str(i): i for i in range(10000)}
    fields = tuple(str(i) for i in range(10000))
    result = extract_dict_columns(data, fields)
    assert result == tuple(range(10000))

def test_extract_dict_columns_partial_missing_fields():
    data = {str(i): i for i in range(10000)}
    fields = tuple(str(i) for i in range(9995, 10005))
    result = extract_dict_columns(data, fields)
    assert result == (9995, 9996, 9997, 9998, 9999, None, None, None, None, None)

def test_extract_dict_columns_mixed_data_types():
    data = {'a': 1, 'b': "text", 'c': 3.14, 'd': [1, 2, 3], 'e': {'key': 'value'}}
    fields = ('a', 'b', 'c', 'd', 'e')
    result = extract_dict_columns(data, fields)
    assert result == (1, "text", 3.14, [1, 2, 3], {'key': 'value'})

def test_extract_dict_columns_repeated_fields():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ('a', 'b', 'a')
    result = extract_dict_columns(data, fields)
    assert result == (1, 2, 1)

def test_extract_dict_columns_nested_structures():
    data = {'a': {'b': {'c': 1}}, 'x': [1, 2, 3], 'y': (4, 5, 6)}
    fields = ('a', 'x', 'y')
    result = extract_dict_columns(data, fields)
    assert result == ({'b': {'c': 1}}, [1, 2, 3], (4, 5, 6))

def test_extract_dict_columns_with_none_values():
    data = {'a': None, 'b': 2, 'c': None}
    fields = ('a', 'b', 'c')
    result = extract_dict_columns(data, fields)
    assert result == (None, 2, None)

def test_extract_dict_columns_empty_data_and_fields():
    data = {}
    fields = ()
    result = extract_dict_columns(data, fields)
    assert result == ()

def test_extract_dict_columns_unicode_keys():
    data = {'α': 1, 'β': 2, 'γ': 3}
    fields = ('α', 'β', 'γ', 'δ')
    result = extract_dict_columns(data, fields)
    assert result == (1, 2, 3, None)

def test_extract_dict_columns_special_characters():
    data = {'!': 1, '@': 2, '#': 3}
    fields = ('!', '@', '#')
    result = extract_dict_columns(data, fields)
    assert result == (1, 2, 3)

def test_extract_dict_columns_very_large_values():
    large_value = "x" * 10**6
    data = {'a': large_value, 'b': large_value, 'c': large_value}
    fields = ('a', 'b', 'c')
    result = extract_dict_columns(data, fields)
    assert result == (large_value, large_value, large_value)

def test_extract_dict_columns_order_consistency():
    data = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
    fields = ('a', 'b', 'c', 'd')

    # Randomize the dictionary's order multiple times
    for _ in range(100):
        items = list(data.items())
        random.shuffle(items)
        randomized_data = dict(items)
        
        result = extract_dict_columns(randomized_data, fields)
        assert result == (1, 2, 3, 4), f"Failed with data: {randomized_data}"

def test_extract_dict_columns_partial_fields_order_consistency():
    data = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
    fields = ('d', 'b', 'a')

    # Randomize the dictionary's order multiple times
    for _ in range(100):
        items = list(data.items())
        random.shuffle(items)
        randomized_data = dict(items)
        
        result = extract_dict_columns(randomized_data, fields)
        assert result == (4, 2, 1), f"Failed with data: {randomized_data}"

def test_extract_dict_columns_randomized_order_with_missing_keys():
    data = {'a': 1, 'b': 2, 'c': 3}
    fields = ('c', 'a', 'd', 'b')

    # Randomize the dictionary's order multiple times
    for _ in range(100):
        items = list(data.items())
        random.shuffle(items)
        randomized_data = dict(items)
        
        result = extract_dict_columns(randomized_data, fields)
        assert result == (3, 1, None, 2), f"Failed with data: {randomized_data}"

def test_extract_dict_columns_empty_dict_with_random_fields():
    data = {}
    fields = ('a', 'b', 'c')

    result = extract_dict_columns(data, fields)
    assert result == (None, None, None)
    
    # Even with different orders in fields, the result should still be None for each missing key
    randomized_fields = list(fields)
    random.shuffle(randomized_fields)
    result = extract_dict_columns(data, tuple(randomized_fields))
    assert result == (None, None, None), f"Failed with fields: {randomized_fields}"

def test_extract_dict_columns_with_same_field_order():
    data = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
    fields1 = ('a', 'b', 'c', 'd')
    fields2 = ('d', 'c', 'b', 'a')

    result1 = extract_dict_columns(data, fields1)
    result2 = extract_dict_columns(data, fields2)
    
    assert result1 == (1, 2, 3, 4)
    assert result2 == (4, 3, 2, 1)

def test_extract_dict_columns_sparse_dict():
    data = {'a': 1}
    fields = ('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i')
    result = extract_dict_columns(data, fields)
    assert result == (1, None, None, None, None, None, None, None, None)

def test_extract_dict_columns_immutability():
    data = {'a': 1, 'b': 2, 'c': 3}
    original_data = data.copy()
    fields = ('a', 'b', 'c')
    extract_dict_columns(data, fields)
    assert data == original_data, "Function modified input data!"

def test_extract_dict_columns_stress_test():
    data = {str(i): i for i in range(10000)}
    fields = tuple(str(i) for i in range(20000))
    result = extract_dict_columns(data, fields)
    assert result[:10000] == tuple(range(10000)) and result[10000:] == (None,) * 10000

def test_extract_dict_columns_duplicate_missing_fields():
    data = {'a': 1, 'b': 2}
    fields = ('a', 'x', 'x', 'b')
    result = extract_dict_columns(data, fields)
    assert result == (1, None, None, 2)

def test_extract_dict_columns_case_sensitivity():
    data = {'A': 1, 'b': 2}
    fields = ('a', 'b')
    result = extract_dict_columns(data, fields)
    assert result == (None, 2)

def test_extract_dict_columns_whitespace_keys():
    data = {' a': 1, 'b ': 2}
    fields = ('a', 'b ')
    result = extract_dict_columns(data, fields)
    assert result == (None, 2)

def test_extract_dict_columns_non_string_keys():
    data = {None: 1, 42: "forty-two"}
    fields = (None, 42, "missing")
    result = extract_dict_columns(data, fields)
    assert result == (1, "forty-two", None)

def test_extract_dict_columns_deeply_nested_keys():
    data = {'a': {'b': {'c': 1}}}
    fields = ('a.b.c', 'a')
    result = extract_dict_columns(data, fields)
    assert result == (None, {'b': {'c': 1}})

def test_extract_dict_columns_large_dataset_with_missing():
    data = {str(i): i for i in range(10**6)}
    fields = ('1000001', '500000', '999999')
    result = extract_dict_columns(data, fields)
    assert result == (None, 500000, 999999)

if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import pyarrow
from orso.types import OrsoTypes

from orso.schema import FlatColumn


def test_column_to_field():
    column = FlatColumn(name="test", type=OrsoTypes.VARCHAR)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.string()

    column = FlatColumn(name="test", type=OrsoTypes.INTEGER)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.int64()

    column = FlatColumn(name="test", type=OrsoTypes.DOUBLE)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.float64()

    column = FlatColumn(name="test", type=OrsoTypes.BOOLEAN)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.bool_()

    column = FlatColumn(name="test", type=OrsoTypes.TIMESTAMP)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.timestamp("us"),  arrow_field.type

    column = FlatColumn(name="test", type=OrsoTypes.DATE)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.date64(), arrow_field.type

    column = FlatColumn(name="test", type=OrsoTypes.BLOB)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.binary()

    column = FlatColumn(name="test", type=OrsoTypes.DECIMAL)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.decimal128(28, 21), arrow_field.type

def test_array_column_to_field():
    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.VARCHAR)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.string())

    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.INTEGER)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.int64())

    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.DOUBLE)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.float64())

    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.BOOLEAN)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.bool_())

    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.TIMESTAMP)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.timestamp("us"))

    column = FlatColumn(name="test", type=OrsoTypes.ARRAY, element_type=OrsoTypes.BLOB)
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.binary())


def test_column_to_field_name():
    column = FlatColumn(name="test", type="ARRAY<VARCHAR>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.string())

    column = FlatColumn(name="test", type="ARRAY<INTEGER>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.int64())

    column = FlatColumn(name="test", type="ARRAY<DOUBLE>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.float64())

    column = FlatColumn(name="test", type="ARRAY<BOOLEAN>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.bool_())

    column = FlatColumn(name="test", type="ARRAY<TIMESTAMP>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.timestamp("us"))

    column = FlatColumn(name="test", type="ARRAY<BLOB>")
    arrow_field = column.arrow_field
    assert arrow_field.type == pyarrow.list_(pyarrow.binary())

def test_field_to_column():
    arrow_field = pyarrow.field("test", pyarrow.list_(pyarrow.int64()))
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.ARRAY
    assert column.element_type == OrsoTypes.INTEGER, column.element_type

    arrow_field = pyarrow.field("test", pyarrow.string())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.VARCHAR
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.int64())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.INTEGER
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.float64())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.DOUBLE
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.bool_())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.BOOLEAN
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.timestamp("us"))
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.TIMESTAMP
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.date32())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.DATE, column.type
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.binary())
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.BLOB
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.decimal128(28, 21))
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.DECIMAL
    assert column.element_type is None

    arrow_field = pyarrow.field("test", pyarrow.list_(pyarrow.string()))
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.ARRAY
    assert column.element_type == OrsoTypes.VARCHAR

    arrow_field = pyarrow.field("test", pyarrow.list_(pyarrow.binary()))
    column = FlatColumn.from_arrow(arrow_field)
    assert column.name == "test"
    assert column.type == OrsoTypes.ARRAY
    assert column.element_type == OrsoTypes.BLOB



if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



def CityHash32(data: bytes) -> int:
    from xxhash import xxh32

    if not isinstance(data, bytes):
        data = str(data).encode()
    return xxh32(data).intdigest()


def CityHash64(data: bytes) -> int:
    from xxhash import xxh64

    if not isinstance(data, bytes):
        data = str(data).encode()
    return xxh64(data).intdigest()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.dataframe import DataFrame
from tests import cities
import re
from typing import List

lengths = {
    0: 5,
    1: 6,
    2: 7,
    3: 8,
    4: 9,
    5: 10,
    6: 11,
    7: 12,
    8: 12,
    9: 12,
    10: 12,
}



def find_all_substrings(s: str, sub: str) -> List[int]:
    """
    Finds all instances of a substring within a string and returns their start indices.

    Parameters:
        s (str): The string to search within.
        sub (str): The substring to search for.

    Returns:
        List[int]: A list of start indices where the substring is found.
    """
    return [m.start() for m in re.finditer(re.escape(sub), s)]

def test_display_ascii_lazy():

    for i in range(10):
        df = DataFrame(cities.values).head(i)
        df._rows = (r for r in df._rows)

        ascii = df.display(limit=3, show_types=True)

        assert len(ascii.split("\n")) == lengths[i], i
        assert len(find_all_substrings(ascii, "Tokyo")) == (1 if i != 0 else 0)


def test_display_ascii_greedy():

    for i in range(10):

        df = DataFrame(cities.values).head(i)
        df.materialize()

        ascii = df.display(limit=3, show_types=True)

        assert len(ascii.split("\n")) == lengths[i], i
        assert len(find_all_substrings(ascii, "Tokyo")) == (1 if i != 0 else 0)



if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# cython: infer_types=True
# cython: embedsignature=True
# cython: binding=False
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cpython.bytes cimport PyBytes_AsString, PyBytes_GET_SIZE
from cpython.object cimport PyObject_Str
from cython cimport int
from datetime import datetime
from ormsgpack import unpackb
from orso.exceptions import DataError
import numpy as np
cimport numpy as cnp
from numpy cimport ndarray
from libc.stdint cimport int32_t, int64_t
from cpython.dict cimport PyDict_GetItem
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM

cnp.import_array()

HEADER_PREFIX = b"\x10\x00"
MAXIMUM_RECORD_SIZE = 8 * 1024 * 1024


cpdef from_bytes_cython(bytes data):
    cdef const char* data_ptr = PyBytes_AsString(data)
    cdef Py_ssize_t length = PyBytes_GET_SIZE(data)

    HEADER_SIZE = 14
    # Validate header and size, now using pointer arithmetic
    if length < HEADER_SIZE or (data_ptr[0] & 0xF0 != 0x10):
        raise DataError("Data malformed")

    # Deserialize record bytes
    cdef Py_ssize_t record_size = (
        (<unsigned char>data_ptr[2]) << 24 |
        (<unsigned char>data_ptr[3]) << 16 |
        (<unsigned char>data_ptr[4]) << 8 |
        (<unsigned char>data_ptr[5])
    )

    if record_size != length - HEADER_SIZE:
        raise DataError("Data malformed - incorrect length")

    # Deserialize and post-process
    cdef list raw_tuple = unpackb(data[HEADER_SIZE:])
    cdef list processed_list = []
    cdef object item

    for item in raw_tuple:
        if isinstance(item, list) and len(item) == 2 and item[0] == "__datetime__":
            processed_list.append(datetime.fromtimestamp(item[1]))
        else:
            processed_list.append(item)

    return tuple(processed_list)

cpdef tuple extract_dict_columns(dict data, tuple fields):
    """
    Extracts the given fields from a dictionary and returns them as a tuple.

    Parameters:
        data: dict
            The dictionary to extract fields from.
        fields: tuple
            The field names to extract.

    Returns:
        A tuple containing values from the dictionary for the requested fields.
        Missing fields will have None.
    """
    cdef int64_t i, num_fields = len(fields)
    cdef void* value_ptr
    cdef list field_data = [None] * num_fields

    for i in range(num_fields):
        value_ptr = PyDict_GetItem(data, fields[i])
        if value_ptr != NULL:
            field_data[i] = <object>value_ptr
        else:
            field_data[i] = None

    return tuple(field_data)  # Convert list to tuple


cpdef cnp.ndarray collect_cython(list rows, int32_t[:] columns, int limit=-1):
    """
    Collects columns from a list of tuples (rows).
    """
    cdef int64_t i, j
    cdef int64_t num_rows = len(rows)
    cdef int64_t num_cols = columns.shape[0]
    cdef int32_t col_idx
    cdef object row
    cdef tuple tuple_row
    
    # Early exit if no rows or columns
    if num_rows == 0 or num_cols == 0:
        return np.empty((num_cols, num_rows), dtype=object)
    
    cdef int64_t row_width = len(rows[0])
    
    # Check if limit is set and within bounds
    if limit >= 0 and limit < num_rows:
        num_rows = limit
    
    # Check if columns are within bounds (only need to check once)
    for j in range(num_cols):
        col_idx = columns[j]
        if col_idx < 0 or col_idx >= row_width:
            raise IndexError(f"Column index out of bounds (0 <= {col_idx} < {row_width})")
    
    # Create result array directly
    cdef cnp.ndarray[object, ndim=2] result = np.empty((num_cols, num_rows), dtype=object)
    
    # Specialized fast paths for common column counts
    if num_cols == 1:
        # Single column case (very common)
        col_idx = columns[0]
        for i in range(num_rows):
            tuple_row = <tuple>rows[i]
            result[0, i] = tuple_row[col_idx]
    elif num_cols == 2:
        # Two column case (also common)
        col_idx0 = columns[0]
        col_idx1 = columns[1]
        for i in range(num_rows):
            tuple_row = <tuple>rows[i]
            result[0, i] = tuple_row[col_idx0]
            result[1, i] = tuple_row[col_idx1]
    else:
        # General case for any number of columns
        for i in range(num_rows):
            tuple_row = <tuple>rows[i]
            for j in range(num_cols):
                result[j, i] = tuple_row[columns[j]]
    
    return result


cpdef int calculate_data_width(cnp.ndarray column_values):
    cdef int width, max_width
    cdef object value

    max_width = 4  # Default width
    for value in column_values:
        if value is not None:
            width = PyBytes_GET_SIZE(PyObject_Str(value))
            if width > max_width:
                max_width = width

    return max_width


from cpython.list cimport PyList_New, PyList_SET_ITEM

def process_table(table, row_factory, int max_chunksize) -> list:
    """
    Processes a PyArrow table and applies a row factory function to each row.

    Parameters:
        table: PyArrow Table
            The input table to process.
        row_factory: function
            A function applied to each row.
        max_chunksize: int
            The batch size to process at a time.

    Returns:
        A list of transformed rows.
    """
    cdef list rows = [None] * table.num_rows
    cdef int64_t i = 0

    for batch in table.to_batches(max_chunksize):
        df = batch.to_pandas().replace({np.nan: None})
        for row in df.itertuples(index=False, name=None):
            rows[i] = row_factory(row)
            i += 1
    return rows



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True

import pyarrow
cimport cython
import struct

from libc.stdint cimport int64_t, uint8_t, int32_t
from libc.stdint cimport int32_t, int64_t, uint8_t, uint64_t, uintptr_t
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM
cpdef list _process_table(table, object row_factory, int max_chunksize):
    """
    Converts a PyArrow table into a list of tuples efficiently.

    Parameters:
        table: PyArrow Table
            The input table to process.
        row_factory: function
            A function applied to each row.
        max_chunksize: int
            The batch size to process at a time.

    Returns:
        A list of transformed rows.
    """
    cdef list result = []
    cdef Py_ssize_t num_cols = table.num_columns
    cdef Py_ssize_t row_idx, col_idx
    cdef object chunk, buffers
    cdef const uint8_t* validity
    cdef const int32_t* int_offsets
    cdef const char* data
    cdef Py_ssize_t str_start, str_end
    cdef bytes value
    cdef tuple row_tuple
    cdef uint8_t null_mask
    cdef Py_ssize_t byte_offset, bit_index
    cdef const uint8_t* raw_data

    for batch in table.to_batches(max_chunksize):
        batch_cols = batch.columns
        batch_num_rows = batch.num_rows

        # Preallocate row storage
        batch_result = [None] * batch_num_rows

        for row_idx in range(batch_num_rows):
            # Create Python list for row data first
            row_data = [None] * num_cols

            for col_idx in range(num_cols):
                chunk = batch_cols[col_idx]
                buffers = chunk.buffers()

                # Extract validity bitmap
                validity = <const uint8_t*><uintptr_t>buffers[0].address if buffers[0] else NULL

                # Compute null mask offsets
                if validity:
                    byte_offset = row_idx // 8
                    bit_index = row_idx % 8
                    null_mask = (validity[byte_offset] >> bit_index) & 1
                else:
                    null_mask = 1  # If no validity buffer, assume all valid

                if null_mask == 0:
                    # NULL value case
                    continue

                # Process based on type
                if pyarrow.types.is_string(chunk.type) or pyarrow.types.is_binary(chunk.type):
                    int_offsets = <const int32_t*><uintptr_t>buffers[1].address
                    data = <const char*><uintptr_t>buffers[2].address if len(buffers) > 2 else NULL

                    str_start = int_offsets[row_idx]
                    str_end = int_offsets[row_idx + 1]
                    
                    if str_start < str_end and data:
                        value = data[str_start:str_end]
                        row_data[col_idx] = value.decode('utf-8', errors='replace')
                    else:
                        row_data[col_idx] = ""

                # Integer handling section - replace with this safer version
                elif pyarrow.types.is_integer(chunk.type):
                    # Get raw pointer to numeric data and safely handle buffer sizes
                    if buffers[1]:
                        raw_data = <const uint8_t*><uintptr_t>buffers[1].address
                        item_size = chunk.type.bit_width // 8
                        buffer_size = buffers[1].size
                        
                        # Calculate offset and make sure we have enough bytes
                        offset = row_idx * item_size
                        if offset + item_size <= buffer_size:
                            if item_size == 8:  # int64
                                row_data[col_idx] = struct.unpack_from("<q", raw_data, offset)[0]
                            elif item_size == 4:  # int32
                                row_data[col_idx] = struct.unpack_from("<i", raw_data, offset)[0]
                            elif item_size == 2:  # int16
                                row_data[col_idx] = struct.unpack_from("<h", raw_data, offset)[0]
                            elif item_size == 1:  # int8
                                row_data[col_idx] = struct.unpack_from("<b", raw_data, offset)[0]
                        else:
                            # Not enough bytes in buffer
                            row_data[col_idx] = None
                    else:
                        row_data[col_idx] = None
                elif pyarrow.types.is_floating(chunk.type):
                    if buffers[1]:
                        raw_data = <const uint8_t*><uintptr_t>buffers[1].address
                        item_size = chunk.type.bit_width // 8
                        buffer_size = buffers[1].size
                        
                        # Calculate offset and make sure we have enough bytes
                        offset = row_idx * item_size
                        if offset + item_size <= buffer_size:
                            if item_size == 8:  # float64
                                row_data[col_idx] = struct.unpack_from("<d", raw_data, offset)[0]
                            elif item_size == 4:  # float32
                                row_data[col_idx] = struct.unpack_from("<f", raw_data, offset)[0]
                        else:
                            row_data[col_idx] = None
                    else:
                        row_data[col_idx] = None
                elif pyarrow.types.is_boolean(chunk.type):
                    # Booleans are bit-packed
                    bool_data = <const uint8_t*><uintptr_t>buffers[1].address if buffers[1] else NULL
                    if bool_data:
                        bool_value = (bool_data[byte_offset] >> bit_index) & 1
                        row_data[col_idx] = bool(bool_value)
                    else:
                        row_data[col_idx] = False

                else:
                    # Fallback for unsupported types
                    try:
                        row_data[col_idx] = chunk[row_idx].as_py()
                    except:
                        row_data[col_idx] = None

            # Convert list to tuple and apply row factory
            row_tuple = tuple(row_data)
            batch_result[row_idx] = row_factory(row_tuple)

        result.extend(batch_result)

    return result


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import decimal
import re
from enum import Enum
from typing import Any
from typing import Iterable
from typing import Tuple
from typing import Type
from typing import Union
from warnings import warn

import orjson

from orso.tools import parse_iso


def _parse_type(type_str: str) -> Union[str, Tuple[str, Tuple[int, ...]]]:
    """
    Parses a SQL type string into its base type and optional parameters.

    Parameters:
        type_str (str): The type definition string (e.g., 'DECIMAL(10,2)', 'VARCHAR[255]', 'ARRAY<VARCHAR>').

    Returns:
        Union[str, Tuple[str, Tuple[int, ...]]]:
            - Just the base type (e.g., "INTEGER", "TEXT").
            - A tuple with the base type and a tuple of integer parameters if applicable (e.g., ("DECIMAL", (10, 2))).
    """

    # Match ARRAY<TYPE>
    array_match = re.match(r"ARRAY<([\w\s\[\]\(\)]+)>", type_str)
    if array_match:
        return "ARRAY", (array_match.group(1),)

    # Match DECIMAL(p,s)
    decimal_match = re.match(r"DECIMAL\((\d+),\s*(\d+)\)", type_str)
    if decimal_match:
        precision, scale = map(int, decimal_match.groups())
        return "DECIMAL", (precision, scale)

    # Match VARCHAR[n]
    varchar_match = re.match(r"VARCHAR\[(\d+)\]", type_str)
    if varchar_match:
        length = int(varchar_match.group(1))
        return "VARCHAR", (length,)

    # Match BLOB[n]
    blob_match = re.match(r"BLOB\[(\d+)\]", type_str)
    if blob_match:
        size = int(blob_match.group(1))
        return "BLOB", (size,)

    # If no parameters, return base type as a string
    return type_str.upper()


class OrsoTypes(str, Enum):
    """
    The names of the types supported by Orso
    """

    ARRAY = "ARRAY"
    BLOB = "BLOB"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    DECIMAL = "DECIMAL"
    DOUBLE = "DOUBLE"
    INTEGER = "INTEGER"
    INTERVAL = "INTERVAL"
    STRUCT = "STRUCT"
    TIMESTAMP = "TIMESTAMP"
    TIME = "TIME"
    VARCHAR = "VARCHAR"
    NULL = "NULL"
    JSONB = "JSONB"
    _MISSING_TYPE = 0

    def __init__(self, *args, **kwargs):
        self._precision: int = None
        self._scale: int = None
        self._element_type: "OrsoTypes" = None
        self._length: int = None

        str.__init__(self)
        Enum.__init__(self)

    def is_numeric(self):
        """is the typle number-based"""
        return self in (self.INTEGER, self.DOUBLE, self.DECIMAL, self.BOOLEAN)

    def is_temporal(self):
        """is the type time-based"""
        return self in (self.DATE, self.TIME, self.TIMESTAMP)

    def is_large_object(self):
        """is the type arbitrary length string"""
        return self in (self.VARCHAR, self.BLOB)

    def is_complex(self):
        return self in (self.ARRAY, self.STRUCT, self.JSONB, self.INTERVAL)

    def __str__(self):
        if self.value == self.ARRAY and self._element_type is not None:
            return f"ARRAY<{self._element_type}>"
        if self.value == self.DECIMAL and self._precision is not None and self._scale is not None:
            return f"DECIMAL({self._precision}, {self._scale})"
        if self.value == self.VARCHAR and self._length is not None:
            return f"VARCHAR[{self._length}]"
        if self.value == self.BLOB and self._length is not None:
            return f"BLOB[{self._length}]"
        return self.value

    def parse(self, value: Any, **kwargs) -> Any:
        if value is None:
            return None
        return ORSO_TO_PYTHON_PARSER[self.value](value, **kwargs)

    @property
    def python_type(self) -> Type:
        return ORSO_TO_PYTHON_MAP.get(self)

    @property
    def numpy_dtype(self):
        import numpy

        MAP = {
            OrsoTypes.ARRAY: numpy.dtype("O"),
            OrsoTypes.BLOB: numpy.dtype("S"),
            OrsoTypes.BOOLEAN: numpy.dtype("?"),
            OrsoTypes.DATE: numpy.dtype("datetime64[D]"),  # [2.5e16 BC, 2.5e16 AD]
            OrsoTypes.DECIMAL: numpy.dtype("O"),
            OrsoTypes.DOUBLE: numpy.dtype("float64"),
            OrsoTypes.INTEGER: numpy.dtype("int64"),
            OrsoTypes.INTERVAL: numpy.dtype("m"),
            OrsoTypes.STRUCT: numpy.dtype("O"),
            OrsoTypes.TIMESTAMP: numpy.dtype("datetime64[us]"),  # [290301 BC, 294241 AD]
            OrsoTypes.TIME: numpy.dtype("O"),
            OrsoTypes.VARCHAR: numpy.dtype("U"),
            OrsoTypes.NULL: numpy.dtype("O"),
        }
        return MAP.get(self)

    @staticmethod
    def from_name(name: str) -> tuple:
        _length = None
        _precision = None
        _scale = None
        _element_type = None

        if name is None:
            return (OrsoTypes._MISSING_TYPE, _length, _precision, _scale, _element_type)

        type_name = str(name).upper()
        parsed_types = _parse_type(type_name)
        if isinstance(parsed_types, str):
            if parsed_types == "ARRAY":
                warn("Column type ARRAY without element_type, defaulting to VARCHAR.")
                _type = OrsoTypes.ARRAY
                _element_type = OrsoTypes.VARCHAR
            elif parsed_types in OrsoTypes.__members__:
                _type = OrsoTypes[parsed_types]
            elif parsed_types == "LIST":
                warn("Column type LIST will be deprecated in a future version, use ARRAY instead.")
                _type = OrsoTypes.ARRAY
            elif parsed_types == "NUMERIC":
                warn(
                    "Column type NUMERIC will be deprecated in a future version, use DECIMAL, DOUBLE or INTEGER instead. Mapped to DOUBLE, this may not be compatible with all values NUMERIC was compatible with."
                )
                _type = OrsoTypes.DOUBLE
            elif parsed_types == "BSON":
                warn("Column type BSON will be deprecated in a future version, use JSONB instead.")
                _type = OrsoTypes.JSONB
            elif parsed_types == "STRING":
                raise ValueError(f"Unknown type '{_type}'. Did you mean 'VARCHAR'?")
            elif (
                type_name == "0"
                or type_name == 0
                or type_name == "VARIANT"
                or type_name == "MISSING"
            ):
                _type = 0
            else:
                raise ValueError(f"Unknown column type '{name}''.")
        elif parsed_types[0] == "ARRAY":
            _type = OrsoTypes.ARRAY
            _element_type = parsed_types[1][0]
            if _element_type.startswith(("ARRAY", "LIST", "NUMERIC", "BSON", "STRING", "DECIMAL")):
                raise ValueError(f"Invalid element type '{_element_type}' for ARRAY type.")
            if _element_type in OrsoTypes.__members__:
                _type = OrsoTypes.ARRAY
                _element_type = OrsoTypes[_element_type]
            else:
                raise ValueError(f"Unknown column type '{_element_type}'.")
        elif parsed_types[0] == "DECIMAL":
            _type = OrsoTypes.DECIMAL
            _precision, _scale = parsed_types[1]
            if _precision < 0 or _precision > 38:
                raise ValueError(f"Invalid precision '{_precision}' for DECIMAL type.")
            if _scale < 0 or _scale > 38:
                raise ValueError(f"Invalid scale '{_scale}' for DECIMAL type.")
            if _precision < _scale:
                raise ValueError(
                    "Precision must be equal to or greater than scale for DECIMAL type."
                )
        elif parsed_types[0] == "VARCHAR":
            _type = OrsoTypes.VARCHAR
            _length = parsed_types[1][0]
        elif parsed_types[0] == "BLOB":
            _type = OrsoTypes.BLOB
            _length = parsed_types[1][0]
        else:
            raise ValueError(f"Unknown column type '{_type}'.")

        return (_type, _length, _precision, _scale, _element_type)


BOOLEAN_STRINGS = (
    "TRUE",
    "ON",
    "YES",
    "1",
    "1.0",
    "T",
    "Y",
    b"TRUE",
    b"ON",
    b"YES",
    b"1",
    b"1.0",
    b"T",
    b"Y",
)


def parse_decimal(value, *, precision=None, scale=None, **kwargs):
    from orso.tools import DecimalFactory

    if value is None:
        return None

    scale = 21 if scale is None else int(scale)
    precision = 38 if precision is None else int(precision)
    value = (
        value.as_py()
        if hasattr(value, "as_py")
        else (
            value.item()
            if hasattr(value, "item") and not isinstance(value, (list, dict, tuple))
            else value
        )
    )
    if isinstance(value, (int, float)):
        value = str(value)
    elif isinstance(value, bytes):
        value = value.decode("utf-8")
    value = value.strip()
    factory = DecimalFactory.new_factory(precision, scale)
    return factory(value)


ORSO_TO_PYTHON_MAP: dict = {
    OrsoTypes.BOOLEAN: bool,
    OrsoTypes.BLOB: bytes,
    OrsoTypes.DATE: datetime.date,
    OrsoTypes.TIMESTAMP: datetime.datetime,
    OrsoTypes.TIME: datetime.time,
    OrsoTypes.INTERVAL: datetime.timedelta,
    OrsoTypes.STRUCT: dict,
    OrsoTypes.DECIMAL: decimal.Decimal,
    OrsoTypes.DOUBLE: float,
    OrsoTypes.INTEGER: int,
    OrsoTypes.ARRAY: list,
    OrsoTypes.VARCHAR: str,
    OrsoTypes.JSONB: bytes,
    OrsoTypes.NULL: None,
}

PYTHON_TO_ORSO_MAP: dict = {
    value: key for key, value in ORSO_TO_PYTHON_MAP.items() if key != OrsoTypes.JSONB
}
PYTHON_TO_ORSO_MAP.update({tuple: OrsoTypes.ARRAY, set: OrsoTypes.ARRAY})  # map other python types


def parse_boolean(x, **kwargs):
    return (x if isinstance(x, (bytes, str)) else str(x)).upper() in BOOLEAN_STRINGS


def parse_bytes(x, **kwargs):
    length = kwargs.get("length")
    if isinstance(x, (dict, list, tuple, set)):
        value = orjson.dumps(x)
    else:
        value = str(x).encode("utf-8") if not isinstance(x, bytes) else x
    if length:
        value = value[:length]
    return value


def parse_date(x, **kwargs):
    result = parse_iso(x)
    if result is None:
        raise ValueError(f"Invalid date.")
    return result.date()


def parse_time(x, **kwargs):
    result = parse_iso(x)
    if result is None:
        raise ValueError(f"Invalid date.")
    return result.time()


def parse_varchar(x, **kwargs):
    length = kwargs.get("length")
    varchar = x.decode("utf-8") if isinstance(x, bytes) else str(x)
    if length:
        varchar = varchar[:length]
    return varchar


def parse_array(x, **kwargs):
    element_type = kwargs.get("element_type")
    if not isinstance(x, (list, tuple, set)):
        x = orjson.loads(x)
    if element_type is None:
        return x
    parser = element_type.parse
    return [parser(v) for v in x]


def parse_double(x, **kwargs):
    return float(x)


def parse_integer(x, **kwargs):
    return int(x)


def parse_null(x, **kwargs):
    return None


def parse_timestamp(x, **kwargs):
    result = parse_iso(x)
    if result is None:
        raise ValueError(f"Invalid timestamp.")
    return result


def parse_interval(x, **kwargs):
    return datetime.timedelta(x)


ORSO_TO_PYTHON_PARSER: dict = {
    OrsoTypes.BOOLEAN: parse_boolean,
    OrsoTypes.BLOB: parse_bytes,
    OrsoTypes.DATE: parse_date,
    OrsoTypes.TIMESTAMP: parse_timestamp,
    OrsoTypes.TIME: parse_time,
    OrsoTypes.INTERVAL: parse_interval,
    OrsoTypes.STRUCT: parse_bytes,
    OrsoTypes.DECIMAL: parse_decimal,
    OrsoTypes.DOUBLE: parse_double,
    OrsoTypes.INTEGER: parse_integer,
    OrsoTypes.ARRAY: parse_array,
    OrsoTypes.VARCHAR: parse_varchar,
    OrsoTypes.JSONB: parse_bytes,
    OrsoTypes.NULL: parse_null,
}


def find_compatible_type(types: Iterable[OrsoTypes], default=OrsoTypes.VARCHAR) -> OrsoTypes:
    """
    Find the most compatible type that can represent all input types.

    Parameters:
        types (list): List of OrsoTypes to find a compatible type for

    Returns:
        OrsoTypes: The most compatible type that can represent all input types

    Examples:
        >>> OrsoTypes.find_compatible_type([OrsoTypes.INTEGER, OrsoTypes.DOUBLE])
        OrsoTypes.DOUBLE
        >>> OrsoTypes.find_compatible_type([OrsoTypes.BLOB, OrsoTypes.VARCHAR])
        OrsoTypes.VARCHAR
    """
    if not types:
        return OrsoTypes.NULL

    # Handle single type case
    if len(set(types)) == 1:
        return types[0]

    # Define type promotion hierarchy
    type_hierarchy = {
        # Numeric promotion
        OrsoTypes.BOOLEAN: 1,
        OrsoTypes.INTEGER: 2,
        OrsoTypes.DOUBLE: 3,
        OrsoTypes.DECIMAL: 4,
        # Temporal promotion
        OrsoTypes.DATE: 1,
        OrsoTypes.TIMESTAMP: 2,
        # String/binary promotion
        OrsoTypes.BLOB: 1,
        OrsoTypes.VARCHAR: 2,
    }

    # First check if all types are in the same category
    if all(t.is_numeric() for t in types):
        return max(types, key=lambda t: type_hierarchy.get(t, 0))
    if all(t.is_temporal() for t in types):
        return max(types, key=lambda t: type_hierarchy.get(t, 0))
    if all(t.is_large_object() for t in types):
        return max(types, key=lambda t: type_hierarchy.get(t, 0))
    if all(
        t in (OrsoTypes.BLOB, OrsoTypes.STRUCT, OrsoTypes.JSONB, OrsoTypes.VARCHAR) for t in types
    ):
        return OrsoTypes.BLOB

    # For heterogeneous types, default to the most flexible type
    if any(t == OrsoTypes.BLOB for t in types):
        return OrsoTypes.BLOB
    return default
